"""Module that defines the FilteringModel class"""
from .base import Module
from torch import Tensor
from .distributions import Distribution

class _obs_from_model(Module):
    def __init__(self, dist:Distribution):
        super().__init__()
        self.dist = dist

    def score(self, state:Tensor, observation:Tensor, **data) -> Tensor:
        return self.dist.log_density(sample = observation.unsqueeze(1), condition_on=state)

    def sample(self, state:Tensor, **data) -> Tensor:
        return self.dist.sample(condition_on = state)

class _dyn_from_model(Module):
    def __init__(self, dist:Distribution):
        super().__init__()
        self.dist = dist

    def log_density(self, prev_state:Tensor, state:Tensor, **data)->Tensor:
        return self.dist.log_density(condition_on=prev_state, sample=state)

    def sample(self, prev_state:Tensor, **data) -> Tensor:
        return self.dist.sample(condition_on=prev_state)

class _prior_from_model(Module):
    def __init__(self, dist:Distribution):
        super().__init__()
        self.dist = dist

    def log_density(self, state:Tensor, **data)->Tensor:
        return self.dist.log_density(sample=state)

    def sample(self, batch_size:int, n_particles:int, **data) -> Tensor:
        return self.dist.sample(sample_size=(batch_size, n_particles))


class FilteringModel(Module):
    r"""The class for grouping model components to define a state-space model with optional proposal distributions.

    Parameters
    ----------
    dynamic_model:Module|Distribution
        The dynamic model.
    observation_model:Module|Distribution
        The observation model.
    prior_model:Module|Distribution
        The prior model.
    initial_proposal_model:Module|Distribution|None. Default None
        The initial proposal model.
    proposal_model:Module|Distribution|None. Default None
        The proposal model.

    Notes
    -----
    Any of the model components can be Distribution objects. However, distributions defined by Distribution objects can only be dependent on at most one variable.
    That variable is:
    For the prior_model and initial_proposal_model: None.
    For the dynamic_model and proposal_model: ``prev_state``.
    For the observation_model: ``state``.

    If the components are not Distributions then they must be Modules with that will be accessed through specific methods. Not all of these methods need to be defined depending on the use case. Let B be the size of the
    batch dimension, K be the size of the particle dimension, D-x be the size of the inherent dimension for data type x. Starred (\*) arguments are always passed, unstarred arguments are only passed if they exist.

    ``prior_model``:

        ``log_density()`` Parameters: \*state, time, control, series_metadata. Output: the probability density of the state, Tensor of size (B X K). Required for non-bootstrap filtering.

        ``sample()`` Parameters: \*batch_size := B, \*n_particles := K, time, control, series_metadata. Output a sample from the prior, Tensor of size (B X K X D-state). Required for data generation and bootstrap filtering.

    ``dynamic_model``:

        ``log_density()`` Parameters: \*prev_state, \*state, prev_time, time, control, series_metadata, \*t. Output: the probability density of the state, Tensor of size (B X K). Required for non-bootstrap filtering.

        ``sample()`` Parameters: \*prev_state, prev_time, time, control, series_metadata, \*t. Output: a sample from the dynamic model, Tensor of size (B X K X D-state). Required for data generation and bootstrap filtering.

    ``observation_model``:

        ``score()`` Parameters: \*state, \*observation, prev_time, time, control, series_metadata, \*t. Output: The score of an observation given the latent state, usually the log-density, Tensor fo size (B X K). Required for filtering.

        ``sample()`` Parameters: \*state, prev_time, time, control, series_metadata, \*t. Output: a sample from the dynamic model, Tensor of size (B X K X D-observation). Required for data generation.

    ``inital_proposal_model``:

        ``log_density()`` Parameters: \*state, \*observation, time, control, series_metadata. Output: the probability density of the state under the initial proposal, Tensor of size (B X K). Required for non-bootstrap filtering.

        ``sample()`` Parameters: \*batch_size := B, \*n_particles := K, \*observation, time, control, series_metadata. Output a sample from the initial proposal, Tensor of size (B X K X D-state). Required non-bootstrap filtering.

    ``proposal_model``:

        ``log_density()`` Parameters: \*prev_state, \*state, \*observation, prev_time, time, control, series_metadata, \*t. Output: the probability density of the state under the proposal, Tensor of size (B X K). Required for non-bootstrap filtering.

        ``sample()`` Parameters: \*prev_state, \*observation, prev_time, time, control, series_metadata, \*t. Output: a sample from the proposal model, Tensor of size (B X K X D-state). Required for non-bootstrap filtering.


    We check that the components have the components required for SIRS particle filtering. Additional components may be required depending on the loss function or when generating data and these will not be caught.
    """
    def __init__(self,
                 *,
                 dynamic_model: Module|Distribution,
                 observation_model: Module|Distribution,
                 prior_model: Module|Distribution,
                 initial_proposal_model: Module|Distribution|None = None,
                 proposal_model: Module|Distribution|None = None):
        super().__init__()
        if isinstance(observation_model, Distribution):
           self.observation_model = _obs_from_model(observation_model)
        else:
            self.observation_model = observation_model
        if isinstance(dynamic_model, Distribution):
            self.dynamic_model = _dyn_from_model(dynamic_model)
        else:
            self.dynamic_model = dynamic_model
        if isinstance(prior_model, Distribution):
            self.prior_model = _prior_from_model(prior_model)
        else:
            self.prior_model = prior_model
        if isinstance(initial_proposal_model, Distribution):
            self.initial_proposal_model = _prior_from_model(initial_proposal_model)
        else:
            self.initial_proposal_model = initial_proposal_model
        if isinstance(proposal_model, Distribution):
            self.proposal_model = _dyn_from_model(proposal_model)
        else:
                self.proposal_model = proposal_model
        if not hasattr(self.observation_model, 'score'):
            raise AttributeError("The observation model must implement a 'score' method")
        
        if self.proposal_model is None:
            if not hasattr(self.dynamic_model, 'sample'):
                raise AttributeError("The dynamic model must implement a 'sample' method")
        else:
            if not hasattr(self.dynamic_model, 'log_density'):
                raise AttributeError("The dynamic model must implement a 'log_density' method")
            if not hasattr(self.proposal_model, 'log_density'):
                raise AttributeError("The proposal model must implement a 'log_density' method")
            if not hasattr(self.proposal_model, 'sample'):
                raise AttributeError("The proposal model must implement a 'sample' method")
            
        if self.initial_proposal_model is None:
            if not hasattr(self.prior_model, 'sample'):
                raise AttributeError("The observation model must implement a 'sample' method")
        else:
            if not hasattr(self.prior_model, 'log_density'):
                raise AttributeError("The prior model must implement a 'log_density' method")
            if not hasattr(self.initial_proposal_model, 'propose'):
                raise AttributeError("The initial sample model must implement a 'sample' method")
            if not hasattr(self.initial_proposal_model, 'log_density'):
                raise AttributeError("The initial proposal model must implement a 'log_density' method")

    @property
    def is_bootstrap(self):
        """True if the model doesn't have either an initial_proposal_model or proposal_model"""
        return self.proposal_model is None and self.initial_proposal_model is None

    @property
    def has_proposal(self):
        """True if the model has a proposal_model"""
        return self.proposal_model is not None

    @property
    def has_initial_proposal(self):
        """True if the model has an initial_proposal_model"""
        return self.initial_proposal_model is not None



from ..base import Module
from abc import ABCMeta, abstractmethod
from typing import Union, Tuple
import torch
from torch import Tensor
from ..utils import multiple_unsqueeze


class Distribution(Module, metaclass=ABCMeta):
    """Base class for implementations of common distributions.
    This is to provide a convenient API, similar to torch.distribution.
    If a user wants to create a custom distribution it will almost always be easier to subclass Module
    and manually implement sample and log_density methods rather than to subclass Distribution.

    Distribution samples have the following dimension order: Batch X Samples X Data

    Data is always a single dimension and is inferred from the distribution parameters, which must not be batched.

    Batch can be any number of batch dimensions, it is inferred from the conditioning variables. For unconditional distributions the batch size
    is always 0.

    Samples can be any number of dimensions, it is manually supplied when calling Distribution.sample(). When calling Distribution.log_density() it
    is inferred as the dimensions of the supplied sample that aren't Batch or Data.

    Log_density returns the log density of each datum in a sample in a batch. Rather than reducing over the sample.
    """

    conditional = False

    def check_sample(self, sample):
        """Check that the sample matches the defined dimension of the distribution."""
        if sample.size(-1) != self.dim:
            raise ValueError(f"Sample must have dimension equal to dimensionality of distribution, found {sample.size(-1)} and {self.dim}")
        if sample.device != sample.device:
            raise ValueError(f"Sample must have device equal to device of distribution parameters, found {sample.device} and {self.device}")

    @staticmethod
    def get_batch_size(size: Tuple[int, ...], data_dims:int) -> Tuple[int, ...]:
        """Get the size of a tensor excluding the last (data_size) dimensions.

        Parameters
        ----------
        size : Tuple[int}
            The size to extract the batch dimensions from.
        data_dims : int
            The number of non-batch dimensions to extract.

        Returns
        -------
        batch_size : Tuple[int]
            The extracted batch dimensions.
        """
        return tuple(list(size)[:-data_dims])

    @staticmethod
    def _unsqueeze_to_size(parameter: Tensor, sample: Union[Tensor, int], data_dims: int = 1) -> Tensor:
        """Unsqueeze a tensor to the same number of dimensions as another.

        Parameters
        ----------
        parameter: Tensor
            The tenosr to unsqueeze.
        sample: Tensor
            The tensor of the size parameter is to be unsqueezed to.
        data_dims: data_dims
            The number of dimensions from the last to unsqueeze at.

        Returns
        -------
        unsqueezed_parameter: Tensor
            parameter unsqueezed to the required size.
        """
        if isinstance(sample, Tensor):
            new_dim = sample.dim()
        else:
            new_dim = sample
        if new_dim - parameter.dim() == 0:
            return parameter
        return multiple_unsqueeze(parameter, new_dim - parameter.dim(), -data_dims-1)

    def __init__(self, generator:torch.Generator, *args, **kwargs) -> None:
        super().__init__()
        self.dim = None
        self.generator = generator
        self.device = self.generator.device
        self.dim = 0

    def __call__(self, *args, **kwargs) -> None:
        # Do not implement a forward method for a Distribution
        pass

    def forward(self, *args, **kwargs) -> None:
        """Do not implement a forward method for a Distribution"""
        pass

    @abstractmethod
    def sample(self, *args, **kwargs) -> Tensor:
        raise NotImplementedError('Sampling not implemented for this distribution')

    @abstractmethod
    def log_density(self, *args, **kwargs) -> Tensor:
        raise NotImplementedError('Density/Mass function not implemented for this distribution')


import torch
from typing import Tuple
from types import FunctionType
from functools import update_wrapper
import os
from warnings import warn
from torch import Tensor


def batched_select(tensor: Tensor, index: torch.LongTensor) -> Tensor:
    """
    Batched analog to tensor[index].
    Use to select along the mth dimension of tensor.
    index has a dimensions , tensor has b dimensions.
    a <= b
    The size of the every dimension but the last of index must be the same as the corresponding dimension in tensor.

    Parameters
    ----------
    tensor : A1 x A2 X ... Am x I x B1 X ... X Bn Tensor
        tensor to select from
    index : A1 x A2 x A3 X ... X Am X D torch.LongTensor
        tensor of indices to select from tensor A1.

    Returns
    -------
    output : A1 x A2 X ... X Am X D X B1 X ... X Bn Tensor
    """
    if tensor.dim() == 3 and index.dim() == 2:
        #Special case common case for efficiency
        return torch.gather(input = tensor, index = index.unsqueeze(-1).expand(-1, -1, tensor.size(2)), dim=1)
    elif tensor.dim() == index.dim():
        return torch.gather(input=tensor, index=index, dim=-1)
    elif tensor.dim() > index.dim():
        index_size = index.size()
        index_dim = index.dim()
        index = multiple_unsqueeze(index, tensor.dim() - index_dim)
        index = index.expand(*index_size,*tuple([tensor.size(i) for i in range(index_dim, tensor.dim())]))
        return torch.gather(input =tensor, index=index, dim=index_dim-1)
    raise ValueError('index cannot have more dimensions than tensor')

def back_index_reduce(tensor: Tensor, index:torch.LongTensor, default_value:torch.Tensor | float = 0., min_entries: int = 0, reduction = 'sum') -> Tensor:
    """
        Acts in the reverse direction to batched_select, applies a reduction over all elements in tensor with the same index.
        Use to select along the mth dimension of tensor.
        index has a dimensions , tensor has b dimensions.
        a <= b
        The size of the every dimension but the last of index must be the same as the corresponding dimension in tensor.

        Parameters
        ----------
        tensor : A1 X A2 X ... Am X D X B1 X B2 X ... X Bn Tensor
            tensor to select from
        index : A1 x A2 x A3 X ... X Am X D torch.LongTensor
            tensor of indices to select from tensor A1.
        default_value : float | Tensor
            Default: 0. The default value to use for indices in output that are not present in the index. If default_value is a Tensor, it's size must be exactly equal to the output size
        min_entries : int
            Default: 0. The size of the indexed dimension in the output will always be the maximum of min_entries and the maximum value in index.
        reduction : str
            Default: 'sum'. The reduction to use, possible options are 'sum', 'prod', 'mean', 'amax', 'amin'.

        Returns
        -------
        output : A1 x A2 X ... X Am X I X B1 X ... X Bn Tensor
    """
    index_dim = index.dim()
    index_size = index.size()
    entries = torch.max(index).item()
    if min_entries is not None:
        entries = max(entries, min_entries)
    if entries != index_size[-1]:
        warn('Warning: backward pass only implemented if index has the same dimension as the output.\nConsider setting min_entries')
    if not isinstance(default_value, torch.Tensor):
        output = torch.full(size=(*tuple(index.size(i) for i in range(0, index_dim-1)), entries, *tuple(tensor.size(j) for j in range(index_dim, tensor.dim()))), fill_value=default_value, dtype=tensor.dtype, device=tensor.device)
    else:
        output = default_value
    index = multiple_unsqueeze(index, tensor.dim() - index_dim)
    index = index.expand(tensor.size())
    return torch.scatter_reduce(output, dim=index_dim-1, index=index, src=tensor, reduce=reduction, include_self=False)



def normalise(tensor: Tensor, dim: int = -1) -> Tuple[Tensor, Tensor]:
    """
    Normalise a log-space tensor to magnitude 1 along a specified dimension.
    Also return the norm.

    Parameters
    ----------
    tensor : Tensor
        tensor to normalise

    dim : int
     dimension to normalise along

    Returns
    -------
    norm_tensor: Tensor
        nomalised tensor

    norm: Tensor
        magnitude of tensor

    """
    norm = torch.logsumexp(tensor, dim=dim, keepdim=True)
    if torch.isinf(norm).any():
        raise ValueError('Row where all values -inf encountered')
    return tensor - norm, norm

def multiple_unsqueeze(tensor: Tensor, n: int, dim: int = -1) -> Tensor:
    """
    Unsqueeze multiple times at the same dimension.
    Equivalent to:

    for i in range(n):
        tensor = tensor.unsqueeze(d)
    return tensor

    Parameters
    ----------
    tensor : Tensor
        Tensor to unsqueeze
    n : int
        Number of times to unsqueeze the tensor
    dim : int
        Dimension to unsqueeze the tensor at

    Returns
    -------
    output : Tensor
        Unsqueezed tensor

    """
    if n == 0:
        return tensor
    if dim < 0:
        dim = tensor.dim() + dim + 1
    return tensor[(slice(None),) * dim + (None, ) * n]


class doc_function:
    """
        Reflection hack to allow functions that only define a docstring.
    """

    def __init__(self, fun):
        """
            Mark an overriding function as docstring only. I.e. the function will default to it's parent's implementation at runtime.
        """
        self.fun = fun
        self.name = fun.__name__
        update_wrapper(self, fun)

    def update_function(self, instance, owner, overrides : FunctionType):
        setattr(owner, self.name, overrides)
        if instance is not None:
            return getattr(instance, self.name)
        #For static methods
        return getattr(owner, self.name)

    def __get__(self, instance, owner):
        if owner is None:
            raise RuntimeError(f'Attempting to use doc_function outside of a class.')
        bases = owner.__mro__
        for base in bases[1:]:
            try:
                overrides = base.__dict__[self.name]
                return self.update_function(instance, owner, overrides)
            except KeyError:
                pass
        raise AttributeError(f'Base classes have no attribute {self.name}.')

def MSE(prediction: Tensor, ground_truth: Tensor):
    return torch.sum(torch.mean((prediction - ground_truth) ** 2, dim=(0,1)))

_deterministic_mode = False

def is_deterministic_mode_enabled():
    return _deterministic_mode

class set_deterministic_mode():
    def __init__(self, mode:bool, warn_only:bool):
        self.prev = is_deterministic_mode_enabled()
        self.mode = mode
        self.warn_only = warn_only
        if mode == True:
            warn('Deterministic mode enabled, this will increase runtime and memory costs and should only be used when reproduciblity is required, reproducibility is only guarenteed on a static hardware and software setup.')


    def set_mode(self):
        os.putenv("CUBLAS_WORKSPACE_CONFIG", ':4096:8')
        torch.use_deterministic_algorithms(True, warn_only=self.warn_only)
        _deterministic_mode = True

    def unset_mode(self):
        os.unsetenv("CUBLAS_WORKSPACE_CONFIG")
        torch.use_deterministic_algorithms(False)
        _deterministic_mode = False

    def __enter__(self):
        if self.prev and (not self.mode):
            self.unset_mode()
        if (not self.prev) and self.mode:
            self.set_mode()

    def __exit__(self, *args):
        if self.prev:
            self.set_mode()
        else:
            self.unset_mode()


from .base import Module
import torch
from torch import Tensor
from typing import Tuple, Callable


class ConditionalResampler(Module):
    def __init__(self, resampler:Module,  condition:Callable):
        """
        A resampling algorithm that conditionally only resamples a subset of the particles from each batch.

        Parameters
        ----------
        resampler : Resampler
            The base resampling algorithm to use.
        condition : Callable
            A function that take the time-step data and returns a 1D boolean tensor where True indicates that resampling should be performed for the given batch and False that it should not.

        Notes
        -----
        We generally do not recommend conditionally resampling in a batched DPF context. Classically, conditional resampling is motivated by computation time, resampling algorithms are typically expensive compared to the
        other components of the filtering loop. However, in a parallelised setting each batch can be resampled in parallel, and it is likely that at least one batch does not satisfy the condition.
        For this reason it is likely that the overhead of calculating and applying the condition will more than cancel out any time savings.
        """
        super().__init__()
        self.resampler = resampler
        self.condition = condition
        self.cache = {}

    def _mask_data(self, mask, **data):
        masked_data = {}
        for k, v in data.items():
            if k == "t":
                masked_data[k] = v
                continue
            masked_data[k] = v[mask]
        return masked_data


    def forward(self, state: Tensor, weight: Tensor, **data) -> tuple[Tensor, Tensor]:
        """Apply the conditional resampler.


        Parameters
        ----------
        state: Tensor
            The particle state
        weight: Tensor
            The particle weight
        data: dict[str, Tensor]
            The remaining data categories

        Returns
        -------
        resampled_state: Tensor
            The conditionally resampled state.
        resampled_weight: Tensor
            The conditionally resampled weight.

        """
        with torch.no_grad():
            resample_mask = self.condition(state=state, weight=weight, **data)
        if not torch.any(resample_mask):
            return state.clone(), weight.clone()
        masked_state = state[resample_mask]
        masked_weight = weight[resample_mask]
        out_state = state.clone()
        out_weight = weight.clone()
        masked_data = self._mask_data(resample_mask, **data)
        resampled_state , resampled_weight = self.resampler(state=masked_state, weight=masked_weight, **masked_data)
        out_state[resample_mask] = resampled_state
        out_weight[resample_mask] = resampled_weight
        self.cache = self.resampler.cache
        t = weight.clone()
        t[resample_mask] = self.cache['used_weight']
        self.cache['used_weight'] = t
        self.cache['mask'] = resample_mask
        return out_state, out_weight

def ESS_Condition(threshold):
    r"""The effective sample size criterion.


    Parameters
    ----------
    threshold: float
        The relative ESS threshold below which resampling should be performed.

    Returns
    -------
    resample_mask:
        A mask tensor that is true if a filter in the batch should be resampled.

    Notes
    -----
    The relative ESS is equal to:

    .. math:: \frac{1}{K\sum^{K}_{k=1}(w^{(k)}_{t})^{2}}

    See [1]_.

    References
    ----------
    .. [1] Elivra V, Martino L, Robert CP (2022), "Rethinking the effective sample size", Int. Stat. Review, 90: 525–550.
    """
    def forward(weight, **data):
        return (1/torch.sum(torch.exp(2*weight), dim=1)) < threshold*weight.size(1)
    return forward


from .datautils import *
from .filtering import *
from .resampling import *
from . import utils
from .base import *
from .distributions import *
from .outputs import *
from .conditional_resampling import *
from .gradient_regularisation import *
from .model_based_api import *
from .Kalman import *


"""This python module contains methods for the Sinkhorn algorithm"""
import torch
from torch import Tensor
from typing import Tuple

def get_transport_from_potentials(log_a: Tensor, log_b: Tensor, cost: Tensor, f: Tensor, g: Tensor, epsilon: float) -> Tensor:
    """Calculates the transport matrix from the Sinkhorn potentials

    Parameters
    ----------
    log_a: (B,M) Tensor
            log of the weights of the proposal distribution
    log_b: (B,N) Tensor
        log of the weights of the target distribution
    cost: (B,M,N) Tensor
        The per unit cost of transporting mass from the proposal to the target
    f: (B,M) pt.Tensor
            Potential on the proposal
    g: (B,N) pt.Tensor
        Potential on the target
    epsilon: float
        Regularising parameter

    Returns
    -------
    T: (B,M,N) Tensor
        The transport matrix
    """
    log_prefactor = log_b.unsqueeze(1) + log_a.unsqueeze(2)
    # Outer sum of f and g
    f_ = torch.unsqueeze(f, 2)
    g_ = torch.unsqueeze(g, 1)
    exponent = (f_ + g_ - cost) / epsilon
    log_transportation_matrix = log_prefactor + exponent
    return torch.exp(log_transportation_matrix)


def apply_transport(x_t: Tensor, transport: Tensor, N: int) -> Tensor:
    """
    Apply a transport matrix to a vector of particles

    Parameters
    ----------
    x_t: (B,N,D) Tensor
        Particle locations to be transported
    transport: (B,M,N) Tensor
        The transport matrix
    N: int
        Number of particles

    """
    return (N * torch.transpose(transport, 1, 2)) @ x_t


def opt_potential(log_a: Tensor, c_potential: Tensor, cost: Tensor, epsilon: Tensor|float) -> Tensor:
    """
        Calculates the update in the Sinkhorn loop for distribution b (either proposal or target)

        Parameters
        ----------
        log_a: (B,N) Tensor
            log of the weights of distribution a
        c_potential: (B, N) Tensor
            the current potential of distribution a
        cost: (B,N,M) Tensor
            The per unit cost of transporting mass from distribution a to distribution b
        epsilon: (B) Tensor|float
            Regularising parameter

        Returns
        -------
        n_potential: (B, M) pt.Tensor
            The updated potential of distribution b

        Notes
        -----
        Sinkhorn algorithm proposed in [1]_.

        References
        ----------
        .. [1] Cuturi M (2013). “Sinkhorn distances: Lightspeed computation of optimal transport.” Proc. Adv. Neural Inf. Process. Syst. (NeurIPS).


    """
    temp = log_a.unsqueeze(1) + (c_potential.unsqueeze(1) - cost) / epsilon
    temp = torch.logsumexp(temp, dim=-1)
    if isinstance(epsilon, Tensor):
        return -epsilon.squeeze(2) * temp
    return -epsilon * temp

def sinkhorn_loop(log_a: Tensor, log_b: Tensor, cost: Tensor, epsilon: float, threshold: float, max_iter: int, state_extent: Tensor, rate: float) -> Tuple[Tensor, Tensor, Tensor]:
    """Calculates the Sinkhorn potentials for entropy regularised optimal transport between two atomic distributions via the Sinkhorn algorithm

        Parameters
        ----------
        log_a: (B,M) Tensor
            log of the weights of the proposal distribution
        log_b: (B,N) Tensor
            log of the weights of the target distribution
        cost: (B,M,N) Tensor
            The per unit cost of transporting mass from the proposal to the target
        epsilon: float
            Regularising parameter
        threshold: float
            The difference in iteratations below which to halt and return
        max_iter: int
            The maximum amount of iterations to run regardless of whether the threshold is hit
        state_extent: (B) Tensor
            The difference between the maximum and minimum value per batch in the state tensor.

        Returns
        -------
        f: (B,M) Tensor
            Potential on the proposal
        g: (B,N) Tensor
            Potential on the target

        Notes
        -----
        Due to convergening to a point, this implementation only retains the gradient at the last step
        Sinkhorn algorithm proposed in [1]_.

        References
        ----------
        .. [1] Cuturi M (2013). “Sinkhorn distances: Lightspeed computation of optimal transport.” Proc. Adv. Neural Inf. Process. Syst. (NeurIPS).
    """
    device = log_a.device
    i = 1
    epsilon_now = torch.clip(state_extent ** 2, min=epsilon).detach()
    cost_T = torch.transpose(cost, 1, 2)
    f_i = opt_potential(log_b, torch.zeros_like(log_b, device=device), cost, epsilon_now)
    g_i = opt_potential(log_a, torch.zeros_like(log_a, device=device), cost_T, epsilon_now)


    continue_criterion = torch.ones((f_i.size(0),), device=device, dtype=torch.bool).unsqueeze(1)

    def stop_criterion(i_, continue_criterion_):
        return i_ < max_iter and torch.any(continue_criterion_)

    #Point convergence, the gradient due to the last step can be substituted for the gradient of the whole loop.
    with torch.no_grad():
        while stop_criterion(i, continue_criterion):
            f_u = torch.where(continue_criterion, (f_i + opt_potential(log_b, g_i, cost, epsilon_now)) / 2, f_i)
            g_u = torch.where(continue_criterion, (g_i + opt_potential(log_a, f_i, cost_T, epsilon_now)) / 2, g_i)
            update_size = torch.maximum(torch.abs(f_u - f_i), torch.abs(g_u - g_i))
            update_size = torch.max(update_size, dim=1)[0]
            continue_criterion = torch.logical_or(update_size > threshold, epsilon_now.squeeze() > epsilon).unsqueeze(1)
            epsilon_now = torch.clip(rate * epsilon_now, min=epsilon)
            f_i = f_u
            g_i = g_u
            i += 1
    f_i = f_i.clone().detach()
    g_i = g_i.clone().detach()
    f = opt_potential(log_b, g_i, cost, epsilon)
    g = opt_potential(log_a, f_i, cost_T, epsilon)
    return f, g, epsilon_now


"""
This module contains the base pydpf.Module class that custom modules for use with pydpf should inherit from.
"""

import torch
from torch.nn import Module as TorchModule
from typing import Any, Callable, Tuple
import functools
from torch import Tensor
import warnings

def _custom_formatwarning(msg, *args, **kwargs):
    return str(msg) + '\n'

warnings.formatwarning = _custom_formatwarning

class DivergenceError(Exception):
    """Custom error type to catch instances of SMC algorithms diverging"""
    pass


class Module(TorchModule):
    """Base class for all modules in PyDPF.
    Includes an update method that should be called after a parameter update to update quantities derived from parameters.
    This is provided to work around pytorch insisting on parameter updates being in place.
    We provide two new function decorators, ``@constrained_parameter`` and ``@cached_property``.
    Both are used to store functions of module parameters that are expensive to compute so it is undesirable to recalculate them everytime
    they are used.
    These provide similar functionality to pytorch's parameterization API, but simpler to use for code that has a lot of custom modules.

    ``@cached_property`` is used to store any intermediate value, for example the inverse of a covariance matrix. Gradient is freely passed through
    the computation of a cached_property.

    ``@constrained_parameter`` should be used to impose constraints only, the underlying data is modified inplace and without gradient. ``@constrained_parameter``
    should always act directly on the underlying parameter, they cannot be stacked or used to constrain functions of parameters.

    Notes
    -----
    A use case not covered by ``@cached_property`` and ``@constrained_parameter``, is if the wrapped function should be computed out-of-place, such as to
    allow gradient tracking, but appear as if the change is made in-place. I.e. we want to modify a parameter from a function declared outside the
    module to which it belongs, and pass gradient through this map. This is the intended use case of the ``torch.parameterization`` API so use that
    instead.
    """

    def __init__(self):
        self.cached_properties = {}
        self.constrained_parameters = {}
        #Need to iterate over __class__.__dict__ rather than dir(self) to bypass getattr()
        for attr, v in self.__class__.__dict__.items():
            if isinstance(v, cached_property):
                self.cached_properties[attr] = v
            if isinstance(v, constrained_parameter):
                self.constrained_parameters[attr] = v
        self.disallow_set_values = False
        super().__init__()


    def forward(self):
        """"""
        raise NotImplementedError('Forward not implemented for this class')

    def __setattr__(self, key: str, value: Any) -> None:

        if key.endswith('_value') and self.disallow_set_values:
            raise AttributeError('Attributes ending with _value are reserved')
        # To be safe if a cached_property is set after object initialisation
        # Not sure how this would come about
        if isinstance(value, cached_property):
            self.cached_properties[key] = value
        if isinstance(value, constrained_parameter):
            self.constrained_parameters[key] = value
        super().__setattr__(key, value)

    def _update(self):
        for name, property in self.cached_properties.items():
            property._update(self)
        for name, property in self.constrained_parameters.items():
            property._update(self)

    def update(self):
        """
        Update all constrained_parameters and cached_properties belonging to this Module.
        """
        self._update()
        for child in self.modules():
            if isinstance(child, Module):
                child._update()


class constrained_parameter:
    """Wrapper for constraining parameters.
    The wrapped function must belong to a Module and should take only a reference to its parent Module and return a reference to the original
    parameter and a tensor containing the new value.

    ``constrained_parameter`` applies the change in-place; the underlying data of the parameter is modified. Necessarily, therefore this is
    done without gradient tracking.

    The constraint is applied on calling ``Module.update()``, so you should do this after every gradient update to preserve the constraint.

    Notes
    -----
    ``constrained_parameters`` can be applied to any parameter accessible from a ``Module`` including as attributes of a child ``Module``.
    This is safe as the parameter is changed in-place.

    .. warning::  ``@constrained_parameter`` should be used to impose constraints only, the underlying data is modified inplace and without gradient. ``@constrained_parameter`` should always act directly on the underlying parameter, they cannot be stacked or used to constrain functions of parameters.
    """

    def __init__(self, function: Callable[[Module], Tuple[Tensor, Tensor]]):
        self.function = function
        functools.update_wrapper(self, function)
        self.value_name = f'{self.function.__name__}_value'

    def __get__(self, instance: Module, owner: Any) -> Tensor:
        if not hasattr(instance, self.value_name):
            self._update(instance)
        return getattr(instance, self.value_name)


    def _update(self, instance: Module) -> None:
        #Changing tensors in place inside inference mode can cause runtime errors.
        @torch.inference_mode(mode=False)
        def f():
            with torch.no_grad():
                d = self.function(instance)
                d[0].data = d[1]
            return d[0]

        v = f()
        instance.disallow_set_values = False
        setattr(instance, self.value_name, v)
        instance.disallow_set_values = True

    def __set__(self):
        raise AttributeError("Cannot directly set a constrained parameter.")


class cached_property:
    """Wrapper for caching functions of parameters.
    The wrapped function must belong to a Module and should take only a reference to its parent Module and tensor containing the new value
    that will be stored.

    cached_property applies its map out-of-place, creating a new tensor. Gradient tracking is permitted through the map.

    The cached_property is calculated lazily; when it is first accessed it is calculated and cached.
    Calling Module.update() resets the value so that it will be recomputed on next access.

    .. warning:: PyTorch generally expects the gradient graph to be created on each forward pass, and destroys it on backwards passes. Therefore, it is recommended to  call ``.update()`` on the top level ``Module`` whenever running a script with gradient tracking even if it is known that the parameters have not changed.
    """
    def __init__(self, function: Callable):
        self.function = function
        functools.update_wrapper(self, function)
        self.value_name = f'{self.function.__name__}_value'

    def __get__(self, instance, owner):
        if not hasattr(instance, self.value_name):
            v = self.function(instance)
            instance.disallow_set_values = False
            setattr(instance, self.value_name, v)
            instance.disallow_set_values = True
        return getattr(instance, self.value_name)

    def _update(self, instance):
        if hasattr(instance, self.value_name):
            delattr(instance, self.value_name)

    def __set__(self, instance):
        raise AttributeError("Cannot directly set a cached property, update the underlying data instead.")


import torch
from ..base import constrained_parameter, cached_property
from .base import Distribution
from torch import Tensor
from torch.distributions.von_mises import VonMises as torch_VM


class VonMises(Distribution):
    """von Mises distribution in radians

    Parameters
    ----------
    mean: Tensor
        The mean of the distribution.
    concentration: Tensor
        The concentration of the distribution. All entries must be positive.
    generator: torch.Generator
        The generator object used to control the RNG.

    Notes
    -----
    This distribution is not reparameterisable as implemented. There is no way to correlate the dimensions of the output sample.
    Each dimension is sampled from a separate von Mises distribution.
    """

    conditional = False

    def __init__(self, mean: Tensor, concentration: Tensor, generator:torch.Generator):
        super().__init__(generator)
        self.mean_ = mean
        self.concentration_ = concentration
        self.generator = generator
        self.dim = mean.size(0)
        self.device = mean.device
        self.torch_dist = torch_VM(self.mean, self.concentration)

    @cached_property
    def _double_concentration(self):
        return self.concentration.to(torch.double)

    @cached_property
    def _double_mean(self):
        return self.mean.to(torch.double)

    @cached_property
    def _proposal_r(self):
        kappa = self._double_concentration
        tau = 1 + (1 + 4 * kappa ** 2).sqrt()
        rho = (tau - (2 * tau).sqrt()) / (2 * kappa)
        _proposal_r = (1 + rho ** 2) / (2 * rho)
        # second order Taylor expansion around 0 for small kappa
        _proposal_r_taylor = 1 / kappa + kappa
        return torch.where(kappa < 1e-5, _proposal_r_taylor, _proposal_r)

    def _rejection_sample(self, x):
        done = torch.zeros(x.shape, dtype=torch.bool, device=self.device)
        while not done.all():
            u = torch.rand((3,) + x.shape, dtype=torch.double, device=self.device, generator=self.generator)
            u1, u2, u3 = u.unbind()
            z = torch.cos(torch.pi * u1)
            f = (1 + self._proposal_r * z) / (self._proposal_r + z)
            c = self._double_concentration * (self._proposal_r - f)
            accept = ((c * (2 - c) - u2) > 0) | ((c / u2).log() + 1 - c >= 0)
            if accept.any():
                x = torch.where(accept, (u3 - 0.5).sign() * f.acos(), x)
                done = done | accept
        return (x + torch.pi + self._double_mean) % (2 * torch.pi) - torch.pi

    @constrained_parameter
    def mean(self):
        mod_angle = torch.remainder(self.mean_, 2*torch.pi)
        return self.mean_, torch.where(mod_angle > torch.pi, mod_angle - 2*torch.pi, mod_angle)

    @constrained_parameter
    def concentration(self):
        return self.concentration_, torch.abs(self.concentration_)

    #The torch implementation has no way to use a generator so we modify their implementation
    @torch.no_grad()
    def sample(self, sample_size: tuple[int,...]|None = None) -> Tensor:
        """Sample a Von-Mises distribution.

        Parameters
        ----------
        sample_size: tuple[int,...]|None
            The size of the sample to draw. Draw a single sample without a sample dimension if None.

        Returns
        -------
        sample: Tensor
            A multivariate Von-Mises sample.
        """
        if sample_size is None:
            true_sample_size = self.mean.size()
        else:
            true_sample_size = (*sample_size, self.mean.size(-1))
        out = torch.empty(true_sample_size, dtype=torch.double, device=self.device)
        return self._rejection_sample(out).to(dtype=self.mean.dtype)

    def log_density(self, sample: Tensor) -> Tensor:
        """Get the density of a Von-Mises distribution.

        pi and -pi are identified.

        Parameters
        ----------
        sample: Tensor
            The sample to get the density of.

        Returns
        -------
        sample log_density: Tensor
            The log density of each datum in the sample.
        """
        return torch.sum(self.torch_dist.log_prob(sample), dim=-1)


from torch import Tensor
import torch
from .base import Distribution
from ..base import Module, constrained_parameter, cached_property
from typing import Tuple, Union, Callable
from ..utils import doc_function
from warnings import warn
from typing import Callable

class MultivariateGaussian(Distribution):
    """An unconditional multivariate Gaussian distribution.

        Parameters
        ----------
        mean: Tensor
            1D tensor specifying the mean.
        cholesky_covariance: Tensor
            2D tensor specifying the (lower) Cholesky decomposition of the covariance matrix. If the upper triangular section has non-zero
            values these will be ignored.
        diagonal_cov: bool
            Whether to constrain the covariance to be diagonal. Default is False.
        generator : Union[torch.Generator, None]
            The generator to control the rng when sampling kernels from the mixture.
    """
    conditional = False

    _half_log2pi = (1/2) * torch.log(torch.tensor(2*torch.pi))

    def __init__(self, mean: Tensor, cholesky_covariance: Tensor, diagonal_cov:bool = False, generator: Union[None, torch.Generator] = None) -> None:


        super().__init__(generator)
        self.mean = mean
        self.cholesky_covariance_ = cholesky_covariance
        self.diagonal_cov = diagonal_cov
        self.dim = mean.size(0)
        if cholesky_covariance.device != mean.device:
            raise ValueError(f'Mean and Covariance should be on the same device, found {mean.device} and {cholesky_covariance.device}')
        if (cholesky_covariance.size(0) != self.dim) or (cholesky_covariance.size(1) != self.dim):
            raise ValueError(f'Covariance must have the same dimensions as the mean, found {self.dim} and {cholesky_covariance.size()}.')

    @constrained_parameter
    def cholesky_covariance(self) -> Tuple[Tensor, Tensor]:
        if self.diagonal_cov:
            diag = torch.diag(self.cholesky_covariance_.diag())
            return  self.cholesky_covariance_,  diag * diag.sign()
        tril = torch.tril(self.cholesky_covariance_)
        diag = tril.diagonal()
        diag.mul_(diag.sign())
        return self.cholesky_covariance_, tril

    @cached_property
    def inv_cholesky_cov(self):
        return torch.linalg.inv_ex(self.cholesky_covariance)[0]

    @cached_property
    def half_log_det_cov(self):
        return torch.linalg.slogdet(self.cholesky_covariance)[1]

    def sample(self, sample_size: tuple[int,...]|None = None) -> Tensor:
        """Sample a Multivariate Gaussian distribution.

        Parameters
        ----------
        sample_size: tuple[int,...]|None
            The size of the sample to draw. Draw a single sample without a sample dimension if None.

        Returns
        -------
        sample: Tensor
            A multivariate Gaussian sample.
        """
        if sample_size is None:
            true_sample_size = self.mean.size()
        else:
            true_sample_size = (*sample_size, self.mean.size(-1))
        output = torch.normal(0, 1, device=self.device, size= true_sample_size, generator=self.generator)
        output = self.mean + output @ self.cholesky_covariance.T
        return output

    def log_density(self, sample: Tensor) -> Tensor:
        """Returns the log density of a sample

        Parameters
        ----------
        sample: Tensor
            The sample to get the density of.

        Returns
        -------
        sample log_density: Tensor
            The log density of each datum in the sample.
        """
        self.check_sample(sample)
        prefactor = -sample.size(-1) * MultivariateGaussian._half_log2pi - self.half_log_det_cov
        residuals = sample - self.mean
        exponent = (-1/2) * torch.sum((residuals @ self.inv_cholesky_cov.T)**2, dim=-1)
        return prefactor + exponent


class StandardGaussian(Distribution):
    """The Multivariate Gaussian distribution with zero mean and diagonal unit covariance. With no learnable parameters.

    Parameters
    ----------
    dim: int
        The dimension of the distribution.
    generator: torch.Generator
        The generator to control the RNG when sampling from this distribution.
    """
    def __new__(cls, dim: int, generator: torch.Generator, learn_mean = False, learn_cov = False, *args, **kwargs):
        cov = torch.eye(dim, device = generator.device)
        mean = torch.zeros(dim, device = generator.device)
        if learn_mean:
            mean = torch.nn.Parameter(mean)
        if learn_cov:
            cov = torch.nn.Parameter(cov)
        return MultivariateGaussian(mean=mean, cholesky_covariance=cov, generator=generator)

class _ConstCovGaussian(Distribution):

    conditional = True

    def _check_conditions(self, condition_on: Tensor) -> None:
        if condition_on.device != self.device:
            raise ValueError(f'condition_on should be on the same device as the distribution parameters, found {condition_on} and {self.device}.')

    def __init__(self, mean: Callable[[Tensor], Tensor], cholesky_covariance: Tensor, diagonal_cov, generator: torch.Generator) -> None:
        super().__init__(generator)
        self.mean_fun = mean
        self.dim = cholesky_covariance.size(0)
        self.dist = MultivariateGaussian(torch.zeros((self.dim,), device=self.device), cholesky_covariance, diagonal_cov, generator)
        self.device = cholesky_covariance.device

    def sample(self, condition_on: Tensor, sample_size: tuple[int,...]|None = None) -> Tensor:
        self._check_conditions(condition_on)
        means = self.mean_fun(condition_on)
        batch_size = self.get_batch_size(means.size(), 1)
        if sample_size is None:
            sample = self.dist.sample(sample_size=batch_size)
        else:
            sample = self.dist.sample(sample_size=(*batch_size, *sample_size))
        return sample + self._unsqueeze_to_size(means, sample)

    def log_density(self, sample: Tensor, condition_on: Tensor) -> Tensor:
        self._check_conditions(condition_on)
        try:
            means = self.mean_fun(condition_on)
        except RuntimeError as e:
            raise RuntimeError(f'Failed to apply condition with error: \n {e}. \n This is likely to a mismatch in batch dimensions between the conditioning variables and the sample.')
        return self.dist.log_density(sample - self._unsqueeze_to_size(means, sample))


class _GeneralCovGaussian(Distribution):
    conditional = True

    def _check_conditions(self, condition_on: Tensor) -> None:
        if condition_on.device != self.device:
            raise ValueError(f'condition_on should be on the same device as the distribution parameters, found {condition_on.device} and {self.device}.')

    def __init__(self,
                mean: Callable[[Tensor], Tensor],
                cholesky_covariance: Callable[[Tensor], Tensor],
                force_diagonal_cov: bool = False,
                dim: Union[int, None] = None,
                device: Union[torch.device, None] = None,
                generator: Union[torch.Generator, None] = None
                ):

        super().__init__(generator)
        self.mean_fun = mean
        self.cov_fun = cholesky_covariance
        self.dim = dim
        self.dist = MultivariateGaussian(torch.zeros((self.dim,), device=self.device), torch.eye(self.dim), True, generator)
        self.device = device
        self.force_diagonal = force_diagonal_cov

    def prepare_cov(self, cov: Tensor) -> Tensor:
        diagonal = torch.abs(torch.diag_embed(torch.diagonal(cov, dim1=-2, dim2=-1)))
        tril = torch.tril(cov, diagonal=-1)
        return diagonal + tril

    def sample(self, condition_on: Tensor, sample_size: tuple[int,...]|None = None) -> Tensor:
        self._check_conditions(condition_on)
        means = self.mean_fun(condition_on)
        cov = self.cov_fun(condition_on)

        batch_size = self.get_batch_size(means.size(), 1)
        if sample_size is None:
            sample = torch.normal(0, 1, device=self.device, size=batch_size, generator=self.generator)
        else:
            sample = torch.normal(0, 1, device=self.device, size=(*batch_size, *sample_size), generator=self.generator)

        means = self._unsqueeze_to_size(means, sample)
        if self.force_diagonal:
            cov = self._unsqueeze_to_size(cov, sample.dim(), 1)
            return sample*cov + means

        cov = self.prepare_cov(cov)
        cov = self._unsqueeze_to_size(cov, sample.dim() + 1, 2)
        return sample.unsqueeze(-2) @ cov  + means

    def log_density(self, sample: Tensor, condition_on: Tensor) -> Tensor:
        self._check_conditions(condition_on)
        means = self.mean_fun(condition_on)
        means = self._unsqueeze_to_size(means, sample)
        cov = self.cov_fun(condition_on)

        residuals = sample - means
        if self.force_diagonal:
            cov = self._unsqueeze_to_size(cov, sample, 1)
            half_log_det_cov = torch.log(torch.prod(cov, dim=-1))
            residuals_cov = residuals/cov
        else:
            cov = self.prepare_cov(cov)
            cov = self._unsqueeze_to_size(cov, sample.dim() + 1, 2)
            half_log_det_cov = torch.linalg.slogdet(cov)[1]
            residuals_cov = torch.linalg.solve_triangular(cov, residuals.unsqueeze(-1))


        prefactor =  -sample.size(-1) * MultivariateGaussian.half_log_2pi - half_log_det_cov

        exponent = (-1 / 2) * torch.sum(residuals_cov ** 2, dim=-1)
        return prefactor + exponent





class ConditionalGaussian(Distribution):
    """A general conditional Gaussian distribution, where the mean and covariance can be given as arbitrary functions of some conditioning tensor.

        Both the mean and cholesky_covariance can be either a Tensor or a function from a Tensor to a Tensor.

        If the mean is a function, it should take an arbitrary conditioning Tensor and output a BxD Tensor where the mean vectors are along the last dimension, and B is zero or more batch dimensions.

        If cholesky_covariance is a fuction and force_diagonal is False, it should take an arbitrary conditioning Tensor and output a BxDxD Tensor where the lower triangular cholesky covariance matricies
        are the last two dimension, and B is zero or more batch dimensions. If the matrix is not a valid lower triangular form it will be mapped as so by setting the diagonal to be positive and all elements above the diagonal to be zero.

        If cholesky_covariance is a fuction and force_diagonal is True, it should take an arbitrary conditioning Tensor and output a BxD Tensor where the last dimension contains the standard deviations of the sample dimensions,
        and B is zero or more batch dimensions.

        Parameters
        ----------
        mean: Callable|Tensor
            The means or a function to calculate them
        cholesky_covariance: Callable|Tensor
            The lower cholesky decomposition of the covariance, or a function to calculate it.
        force_diagonal_cov: bool
            Whether to force the covariance matrix to be diagonal or not.
        dim: int|None
            If both the mean and the covariance are given as functions, then the dimension of the distribution must be provided, otherwise it is inferred from the constant parameters.
        device : torch.device|None
            If both the mean and the covariance are given as functions, then the device of the distribution must be provided, otherwise it is inferred from the constant parameters.
        generator : torch.Generator
            The generator to control the rng when sampling.


        Notes
        -----
        In the most general case, repeatedly computing large batches of matrix determinants and inverses is slow. We implement optimised routines for the special cases that the covariance is constant or that it is diagonal.
        If the user's problem does not fit these cases but has other structure that can be taken advantage of it is recommended that they implement a custom sampling proceedure rather than use this.
        """

    def __new__(cls, mean: Callable|Tensor,
                cholesky_covariance: Callable|Tensor,
                force_diagonal_cov: bool = False,
                dim: Union[int, None] = None,
                device: torch.device|None = None,
                generator: torch.Generator = torch.default_generator
                ) -> Distribution:

        if isinstance(mean, Tensor) and isinstance(cholesky_covariance, Tensor):
            warn('Using the conditional API to create a non-conditional distribution, are you sure this is correct?')
            return MultivariateGaussian(mean, cholesky_covariance, force_diagonal_cov, generator)

        if isinstance(cholesky_covariance, Tensor):
            return _ConstCovGaussian(mean, cholesky_covariance, force_diagonal_cov, generator)

        if isinstance(mean, Tensor):
            return _GeneralCovGaussian(lambda t: mean, cholesky_covariance, force_diagonal_cov, mean.size(-1), mean.device,generator)

        if dim is None:
            raise ValueError('Dimension must be specified if both the mean and the covariance are functions')
        if device is None:
            raise ValueError('Dimension must be specified if both the mean and the covariance are functions')

        return _GeneralCovGaussian(mean, cholesky_covariance, force_diagonal_cov, dim, device, generator)


    @doc_function
    def __init__(self, mean: Union[Callable[[Tensor], Tensor], Tensor],
                cholesky_covariance: Union[Callable[[Tensor], Tensor], Tensor],
                force_diagonal_cov: bool = False,
                dim: Union[int, None] = None,
                device: Union[torch.device, None] = None,
                generator: Union[torch.Generator, None] = None,
                ):


        super().__init__(generator)

    @doc_function
    def sample(self, condition_on: Tensor, sample_size: tuple[int,...]|None = None) -> Tensor:
        """
        Sample a conditional Gaussian distribution.

        Parameters
        ----------
        condition_on: Tensor
            The vector to condition the distribution on.

        sample_size: Union[Tuple[int, ...], None]
            The size of the sample to draw. If None then a single sample is drawn per batch dimension and no sample dimension is used.

        Returns
        -------
        sample: Tensor
            A sample of a conditional Gaussian, conditioned on condition_on.

        """
        pass


    @doc_function
    def log_density(self, sample: Tensor, condition_on: Tensor) -> Tensor:
        """
        Evaluate the log density of a sample.

        Parameters
        ----------
        sample: Tensor
            The sample to get the density of.

        condition_on: Tensor
            The vector to condition the distribution on.

        Returns
        -------
        sample log_density: Tensor
            The log density of each datum in the sample.

        """
        pass

class LinearGaussian(Distribution):
    """A Gaussian conditional distribution, where the means of the Gaussian are conditional on a supplied variable, X, through the linear map :math:`WX + B`. Where W is the weights and B is the bias.

        Parameters
        ----------
        weight: Tensor
            2D tensor specifying the weight matrix, W.
        bias: Tensor
            1D tensor specifying the bias, B.
        cholesky_covariance: Tensor
            2D tensor specifying the (lower) Cholesky decomposition of the covariance matrix. If the upper triangular section has non-zero
            values these will be ignored.
        diagonal_cov: bool
            Whether to force the covariance matrix to be diagonal or not.
        constrain_spectral_radius: Union[int, None]
            If constrain_spectral_radius is an integer, then the weight matrix will be scaled so that it's spectral radius never exceed the passed value. If constrain_spectral_radius is None, no scaling is applied.
        generator : Union[torch.Generator, None]
            The generator to control the rng when sampling kernels from the mixture.
    """

    conditional = True

    def __new__(cls,
                weight: Tensor,
                bias: Tensor,
                cholesky_covariance: Tensor,
                diagonal_cov: bool = False,
                constrain_spectral_radius: Union[float, None] = None,
                generator: Union[torch.Generator, None] = None):

        class LinearTransform(Module):
            def __init__(self):
                super().__init__()
                self.weight = weight
                self.bias = bias

            def forward(self, x):
                return (self.constrained_weight @ x.unsqueeze(-1)).squeeze(-1) + self.bias

            @constrained_parameter
            def constrained_weight(self):
                if constrain_spectral_radius is not None:
                    eigvals = torch.linalg.eigvals(self.weight)
                    spectral_radius = torch.max(torch.abs(eigvals))
                    if  spectral_radius > constrain_spectral_radius:
                        return self.weight, self.weight / spectral_radius
                return self.weight, self.weight

        dim = weight.size(-2)
        device = weight.device
        if (cholesky_covariance.size(0) != dim) or (cholesky_covariance.size(1) != dim):
            raise ValueError(f'Covariance must have the same dimensions as the weights first dimension, found {cholesky_covariance.size()} and {weight.size()}.')
        if bias.size(0) != dim:
            raise ValueError(f'Bias must have the same dimensions as the weights first dimension, found {bias.size()} and {weight.size()}')
        if cholesky_covariance.device != device:
            raise ValueError(f'Weight and Covariance should be on the same device, found {device} and {cholesky_covariance.device}')
        if cholesky_covariance.device != device:
            raise ValueError(f'Weight and bias should be on the same device, found {device} and {bias.device}')

        return _ConstCovGaussian(LinearTransform(), cholesky_covariance, diagonal_cov, generator)

    def sample(self, condition_on: Tensor, sample_size: tuple[int,...]|None = None) -> Tensor:
        """
        Sample a multivariate Linear Gaussian.
        The means of the Gaussian are calculated as condition_on @ self.weight + self.bias.

        Parameters
        ----------
        condition_on: Tensor
            The vector to condition the distribution on.
        sample_size: Union[Tuple[int, ...], None]
            The size of the sample to draw. If None then a single sample is drawn per batch dimension and no sample dimension is used.

        Returns
        -------
        sample: Tensor
            A sample of a multivariate Linear Gaussian, conditioned on condition_on.

        """
        pass

    def log_density(self, sample: Tensor, condition_on: Tensor) -> Tensor:
        """
        Evaluate the log density of a sample.
        The means of the Gaussian are calculated as condition_on @ self.weight + self.bias.

        Parameters
        ----------
        sample: Tensor
            The sample to get the density of.
        condition_on: Tensor
            The vector to condition the distribution on.

        Returns
        -------
        sample log_density: Tensor
            The log density of each datum in the sample.

        """
        pass